from collections import defaultdict
from collections.abc import Callable
from typing import override

from greyhorse.app.abc.collectors import (
    Collector,
    MutCollector,
    MutNamedCollector,
    NamedCollector,
)
from greyhorse.app.abc.selectors import ListSelector, NamedListSelector
from greyhorse.maybe import Just, Maybe, Nothing


class DictRegistry[K, T](Collector[K, T], ListSelector[K, T]):
    __slots__ = ('_storage',)

    def __init__(self) -> None:
        self._storage: dict[K, T] = {}

    def __len__(self) -> int:
        return len(self._storage)

    @override
    def add(self, key: K, instance: T) -> bool:
        if key in self._storage:
            return instance is self._storage[key]
        self._storage[key] = instance
        return True

    @override
    def has(self, key: K) -> bool:
        return key in self._storage

    @override
    def get(self, key: K) -> Maybe[T]:
        if key not in self._storage:
            return Nothing

        return Just(self._storage[key])

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]:
        result = []

        for k, v in self._storage.items():
            if not filter_fn or filter_fn(k):
                result.append((k, v))

        return result


class MutDictRegistry[K, T](MutCollector[K, T], DictRegistry[K, T]):
    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        if key not in self._storage:
            return False

        value = self._storage[key]

        if instance is None:
            del self._storage[key]

        else:
            if instance != value:
                return False
            del self._storage[key]

        return True


class ScopedDictRegistry[K, T](DictRegistry[K, T]):
    def __init__(
        self, factory: Callable[[], DictRegistry[K, T]], scope_func: Callable[[], str]
    ) -> None:
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, DictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> DictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    def __len__(self) -> int:
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]:
        registry = self._get_registry()
        return registry.items(filter_fn)


class ScopedMutDictRegistry[K, T](MutDictRegistry[K, T]):
    def __init__(
        self, factory: Callable[[], MutDictRegistry[K, T]], scope_func: Callable[[], str]
    ) -> None:
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, MutDictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> MutDictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    def __len__(self) -> int:
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]:
        registry = self._get_registry()
        return registry.items(filter_fn)

    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        registry = self._get_registry()
        return registry.remove(key, instance)


class NamedDictRegistry[K, T](NamedCollector[K, T], NamedListSelector[K, T]):
    __slots__ = ('_storage',)

    def __init__(self) -> None:
        self._storage: dict[K, dict[str | None, T]] = defaultdict(dict)

    def __len__(self) -> int:
        return len(self._storage)

    @override
    def add(self, key: K, instance: T, name: str | None = None) -> bool:
        if key in self._storage and name in self._storage[key]:
            return instance is self._storage[key][name]
        self._storage[key][name] = instance
        return True

    @override
    def has(self, key: K, name: str | None = None) -> bool:
        if key not in self._storage:
            return False
        if name is None:
            return True
        return name in self._storage[key]

    @override
    def get(self, key: K, name: str | None = None) -> Maybe[T]:
        if key not in self._storage:
            return Nothing

        if name is None:
            first_name = next(iter(self._storage[key]))
            return Just(self._storage[key][first_name])
        return Just(self._storage[key][name])

    @override
    def items(
        self, filter_fn: Callable[[K, str], bool] | None = None
    ) -> list[tuple[K, str, T]]:
        result = []

        for k, values in self._storage.items():
            for name, v in values.items():
                if not filter_fn or filter_fn(k, name):
                    result.append((k, name, v))

        return result


class MutNamedDictRegistry[K, T](MutNamedCollector[K, T], NamedDictRegistry[K, T]):
    @override
    def remove(self, key: K, instance: T | None = None, name: str | None = None) -> bool:
        if key not in self._storage:
            return False

        if name is None:
            values = self._storage[key]

            if instance is None:
                del self._storage[key]
            else:
                while len(values):
                    name = next(iter(values))
                    if instance == values[name]:
                        del values[name]
                if len(values) == 0:
                    del self._storage[key]

            return True

        if name not in self._storage[key]:
            return False

        del self._storage[key][name]

        if len(self._storage[key]) == 0:
            del self._storage[key]
        return True



from typing import TYPE_CHECKING, Any

from greyhorse.app.abc.controllers import Controller, ControllerError, ControllerFactoryFn
from greyhorse.app.abc.providers import Provider
from greyhorse.app.abc.services import Service, ServiceError, ServiceFactoryFn
from greyhorse.app.registries import MutDictRegistry, MutNamedDictRegistry
from greyhorse.app.schemas.components import ComponentConf, ModuleComponentConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.result import Err, Ok, Result
from greyhorse.utils.injectors import ParamsInjector

from ...maybe import Just, Maybe, Nothing
from ...utils.invoke import invoke_sync
from ..abc.selectors import NamedListSelector, NamedSelector
from ..abc.visitor import Visitor
from ..private.res_manager import ResourceManager

if TYPE_CHECKING:
    from .module import Module


class ComponentError(Error):
    namespace = 'greyhorse.app.component'

    Resource = ErrorCase(
        msg='{path}: Resource error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Ctrl = ErrorCase(
        msg='{path}: Controller error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Service = ErrorCase(
        msg='{path}: Service error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Module = ErrorCase(
        msg='{path}: Submodule error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )


class Component:
    def __init__(self, name: str, path: str, conf: ComponentConf) -> None:
        self._name = name
        self._path = path
        self._conf = conf
        self._rm = ResourceManager()

        self._controllers: list[Controller] = []
        self._services: list[Service] = []

        self._resources = MutNamedDictRegistry[type, Any]()
        self._providers = MutDictRegistry[type[Provider], Provider]()

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_component(self)
        for svc in self._services:
            svc.accept_visitor(visitor)
        for ctrl in self._controllers:
            ctrl.accept_visitor(visitor)
        visitor.finish_component(self)

    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]:
        return self._rm.find_provider(prov_type, self._providers).map(Just).unwrap_or(Nothing)

    def add_provider[T](self, prov_type: type[Provider[T]], provider: Provider[T]) -> bool:
        return self._providers.add(prov_type, provider)

    def remove_provider[T](self, prov_type: type[Provider[T]]) -> bool:
        return self._providers.remove(prov_type)

    def add_resource(self, res_type: type, resource: Any, name: str | None = None) -> bool:
        return self._resources.add(res_type, resource, name=name)

    def remove_resource(self, res_type: type, name: str | None = None) -> bool:
        return self._resources.remove(res_type, name=name)

    def add_controller(self, controller: Controller) -> bool:
        if self._controllers.count(controller):
            return False
        self._controllers.append(controller)
        return self._rm.add_controller(controller)

    def remove_controller(self, controller: Controller) -> bool:
        if not self._controllers.count(controller):
            return False
        self._controllers.remove(controller)
        return self._rm.remove_controller(controller)

    def add_service(self, service: Service) -> bool:
        if self._services.count(service):
            return False
        self._services.append(service)
        return self._rm.add_service(service)

    def remove_service(self, service: Service) -> bool:
        if not self._services.count(service):
            return False
        self._services.remove(service)
        return self._rm.remove_service(service)

    def setup(self) -> Result[None, ComponentError]:
        injector = ParamsInjector()

        logger.info('{path}: Component "{name}" setup'.format(path=self._path, name=self.name))

        if not (res := self._create_services(injector)):
            return res

        for svc in res.unwrap():
            injector.add_type_provider(type(svc), svc)
            self.add_service(svc)

        if not (res := self._create_controllers(injector)):
            return res

        for svc in self._services:
            injector.remove_type_provider(type(svc))

        for ctrl in res.unwrap():
            self.add_controller(ctrl)

        if not (
            res := self._rm.setup(self._providers).map_err(
                lambda e: ComponentError.Resource(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        for ctrl, _ctrl_conf in zip(self._controllers, self._conf.controllers, strict=False):
            if not (
                res := invoke_sync(ctrl.setup, self._resources).map_err(
                    lambda e: ComponentError.Ctrl(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

        injector.add_type_provider(NamedSelector[type, Any], self._resources)
        injector.add_type_provider(NamedListSelector[type, Any], self._resources)

        for svc, svc_conf in zip(self._services, self._conf.services, strict=False):
            for res_type in svc_conf.resources:
                injector.add_type_provider(Maybe[res_type], self._resources.get(res_type))

            injected_args = injector(svc.setup)

            if not (
                res := invoke_sync(
                    svc.setup, *injected_args.args, **injected_args.kwargs
                ).map_err(
                    lambda e: ComponentError.Service(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

            for res_type in svc_conf.resources:
                injector.remove_type_provider(Maybe[res_type])

        injector.remove_type_provider(NamedSelector[type, Any])
        injector.remove_type_provider(NamedListSelector[type, Any])

        logger.info(
            '{path}: Component "{name}" setup successful'.format(
                path=self._path, name=self.name
            )
        )

        return Ok()

    def teardown(self) -> Result[None, ComponentError]:
        injector = ParamsInjector()

        logger.info(
            '{path}: Component "{name}" teardown'.format(path=self._path, name=self.name)
        )

        injector.add_type_provider(NamedSelector[type, Any], self._resources)
        injector.add_type_provider(NamedListSelector[type, Any], self._resources)

        for svc, svc_conf in zip(
            reversed(self._services), reversed(self._conf.services), strict=False
        ):
            for res_type in svc_conf.resources:
                injector.add_type_provider(Maybe[res_type], self._resources.get(res_type))

            injected_args = injector(svc.setup)

            if not (
                res := invoke_sync(
                    svc.teardown, *injected_args.args, **injected_args.kwargs
                ).map_err(
                    lambda e: ComponentError.Service(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

            for res_type in svc_conf.resources:
                injector.remove_type_provider(Maybe[res_type])

        injector.remove_type_provider(NamedSelector[type, Any])
        injector.remove_type_provider(NamedListSelector[type, Any])

        for ctrl, _ctrl_conf in zip(
            reversed(self._controllers), reversed(self._conf.controllers), strict=False
        ):
            if not (
                res := invoke_sync(ctrl.teardown, self._resources).map_err(
                    lambda e: ComponentError.Ctrl(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

        if not (
            res := self._rm.teardown().map_err(
                lambda e: ComponentError.Resource(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        while ctrl := next(reversed(self._controllers), None):
            self.remove_controller(ctrl)

        while svc := next(reversed(self._services), None):
            self.remove_service(svc)

        assert not self._controllers
        assert not self._services

        logger.info(
            '{path}: Component "{name}" teardown successful'.format(
                path=self._path, name=self.name
            )
        )

        return Ok()

    def _create_controller(
        self, conf: CtrlConf, factory: ControllerFactoryFn, injector: ParamsInjector
    ) -> Result[Controller, ControllerError]:
        logger.info(
            '{path}: Component controller "{name}" create'.format(
                path=self._path, name=conf.name
            )
        )

        values = conf.args.copy()
        values['name'] = conf.name
        injected_args = injector(factory, values=values)

        try:
            if not (res := factory(*injected_args.args, **injected_args.kwargs)):
                return res

        except Exception as e:
            error = ControllerError.Factory(details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component controller "{name}" created successfully'.format(
                path=self._path, name=conf.name
            )
        )

        if isinstance(res, Controller):
            return Ok(res)

        return res

    def _create_service(
        self, conf: SvcConf, factory: ServiceFactoryFn, injector: ParamsInjector
    ) -> Result[Service, ServiceError]:
        logger.info(
            '{path}: Component service "{name}" create'.format(path=self._path, name=conf.name)
        )

        values = conf.args.copy()
        values['name'] = conf.name
        injected_args = injector(factory, values=values)

        try:
            if not (res := factory(*injected_args.args, **injected_args.kwargs)):
                return res

        except Exception as e:
            error = ServiceError.Factory(details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component service "{name}" created successfully'.format(
                path=self._path, name=conf.name
            )
        )

        if isinstance(res, Service):
            return Ok(res)

        return res

    def _create_controllers(
        self, injector: ParamsInjector
    ) -> Result[list[Controller], ComponentError]:
        result = []

        for conf in self._conf.controllers:
            factory = self._conf.controller_factories.get(conf.type_, conf.type_)

            match self._create_controller(conf, factory, injector):
                case Ok(ctrl):
                    result.append(ctrl)

                case Err(e):
                    error = ComponentError.Ctrl(
                        path=self._path, name=self.name, details=e.message
                    )
                    logger.error(error.message)
                    return error.to_result()

        return Ok(result)

    def _create_services(
        self, injector: ParamsInjector
    ) -> Result[list[Service], ComponentError]:
        result = []

        for conf in self._conf.services:
            factory = self._conf.service_factories.get(conf.type_, conf.type_)

            match self._create_service(conf, factory, injector):
                case Ok(svc):
                    result.append(svc)

                case Err(e):
                    error = ComponentError.Service(
                        path=self._path, name=self.name, details=e.message
                    )
                    logger.error(error.message)
                    return error.to_result()

        return Ok(result)


class ModuleComponent(Component):
    def __init__(
        self, name: str, path: str, conf: ModuleComponentConf, module: 'Module'
    ) -> None:
        super().__init__(name, path, conf)
        self._conf = conf
        self._module = module

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_component(self)
        for svc in self._services:
            svc.accept_visitor(visitor)
        for ctrl in self._controllers:
            ctrl.accept_visitor(visitor)
        self._module.accept_visitor(visitor)
        visitor.finish_component(self)

    def setup(self) -> Result[None, ComponentError]:
        if not (res := super().setup()):
            return res

        for op in self._rm.get_operators():
            self._module.add_operator(op)

        for res_type, res_name, res in self._resources.items():
            self._module.add_resource(res_type, res, res_name)

        for prov_type, prov in self._providers.items():
            self._module.add_provider(prov_type, prov)

        if not (
            res := self._module.setup().map_err(
                lambda e: ComponentError.Module(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        return Ok()

    def teardown(self) -> Result[None, ComponentError]:
        if not (
            res := self._module.teardown().map_err(
                lambda e: ComponentError.Module(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        for prov_type, _ in self._providers.items():  # noqa: PERF102
            self._module.remove_provider(prov_type)

        for res_type, res_name, _ in self._resources.items():
            self._module.remove_resource(res_type, res_name)

        for op in self._rm.get_operators():
            self._module.remove_operator(op)

        return super().teardown()



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any, Generic, Self, TypeVar

IdType = TypeVar('IdType')


class AbstractModel(Generic[IdType], ABC):
    @abstractmethod
    def get_id_value(self) -> IdType: ...

    #
    # Retrieve operations
    #

    @classmethod
    @abstractmethod
    async def get(cls, id_value: IdType, **kwargs) -> Self | None: ...

    @classmethod
    @abstractmethod
    async def get_any(cls, indices: Sequence[IdType], **kwargs) -> Sequence[Self | None]: ...

    @classmethod
    @abstractmethod
    async def exists(cls, id_value: IdType, **kwargs) -> bool: ...

    @abstractmethod
    async def load(self, only: Sequence[str] | None = None) -> bool: ...

    #
    # Modification operations
    #

    @classmethod
    @abstractmethod
    async def create(cls, data: Mapping[str, Any], **kwargs) -> Self | None: ...

    @classmethod
    @abstractmethod
    async def get_or_create(
        cls, id_value: IdType, data: Mapping[str, Any], **kwargs
    ) -> tuple[Self | None, bool]: ...

    @abstractmethod
    async def update(self, data: Mapping[str, Any], **kwargs) -> bool: ...

    @classmethod
    @abstractmethod
    async def update_by_id(
        cls, id_value: IdType, data: Mapping[str, Any], **kwargs
    ) -> bool: ...

    @abstractmethod
    async def save(self, **kwargs) -> bool: ...

    @classmethod
    @abstractmethod
    async def save_all(cls, objects: Sequence[Self], **kwargs) -> bool: ...

    @abstractmethod
    async def delete(self) -> bool: ...

    @classmethod
    @abstractmethod
    async def delete_all(cls, indices: Sequence[IdType] | None = None) -> bool: ...

    @classmethod
    @abstractmethod
    async def delete_by_id(cls, id_value: IdType) -> bool: ...



import importlib.metadata
import inspect
from pathlib import Path


def get_project_path() -> Path:
    for frame in reversed(inspect.stack()):
        path = Path(frame.filename).absolute()

        while path.parent != path:
            path = path.parent
            pyproject_toml_path = path / 'pyproject.toml'
            if pyproject_toml_path.exists():
                return pyproject_toml_path

    raise AssertionError('Could not find pyproject.toml')


def get_version() -> str:
    from tomlkit import parse

    pyproject_toml_path = get_project_path()

    with pyproject_toml_path.open('r') as f:
        pyproject_toml = parse(string=f.read())

    project = pyproject_toml['project']
    return importlib.metadata.version(project['name'])



import logging.config
import os
import sys

DEBUG_MODE = os.environ.get('DEBUG', False)


LOGGING_CONFIG_DEFAULTS = dict(
    version=1,
    disable_existing_loggers=False,
    loggers={
        'greyhorse': {
            'level': 'INFO' if not DEBUG_MODE else 'DEBUG',
            'handlers': ['console'],
            'propagate': False,
        },
        'greyhorse.wtf': {
            'level': 'INFO',
            'handlers': ['error_console'],
            'propagate': True,
            'qualname': 'greyhorse.wtf',
        },
    },
    handlers={
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
            'stream': sys.stdout,
        },
        'error_console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
            'stream': sys.stderr,
        },
    },
    formatters={
        'generic': {
            'format': '%(asctime)s [%(process)d] [%(levelname)s] %(message)s',
            'datefmt': '[%Y-%m-%d %H:%M:%S %z]',
            'class': 'logging.Formatter',
        },
        'verbose': {
            'format': '%(asctime)s [%(process)d/%(threadName)s] [%(levelname)-5.8s] %(message)s (%(name)s:%(module)s:%(lineno)s)',
            'datefmt': '[%Y-%m-%d %H:%M:%S %z]',
            'class': 'logging.Formatter',
        },
        'simple': {'format': '%(asctime)s [%(levelname)-5.8s] %(message)s'},
    },
)


logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)
logger = logging.getLogger('greyhorse')
logger_wtf = logging.getLogger('greyhorse.wtf')



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Iterable, Mapping
from typing import Any, Protocol

from greyhorse.data.query import Query
from greyhorse.error import Error, ErrorCase
from greyhorse.maybe import Maybe
from greyhorse.result import Ok, Result
from greyhorse.utils.types import TypeWrapper


class EntityError(Error):
    namespace = 'greyhorse.data'

    Validation = ErrorCase(msg='Entity validation error: "{details}"', details=str)
    AlreadyExists = ErrorCase(msg='Entity already exists')
    NotOnlyOne = ErrorCase(msg='The result is not the only one')
    Empty = ErrorCase(msg='The result is empty')
    Deleted = ErrorCase(msg='Cannot perform operation because the entity is deleted')


class Repository[E, ID](Protocol):
    @abstractmethod
    def get(self, id_value: ID, **kwargs) -> Maybe[E] | Awaitable[Maybe[E]]: ...

    @abstractmethod
    def get_any(
        self, indices: Iterable[ID], **kwargs
    ) -> Iterable[Maybe[E]] | Awaitable[Iterable[Maybe[E]]]: ...

    @abstractmethod
    def exists(self, id_value: ID, **kwargs) -> bool | Awaitable[bool]: ...

    @abstractmethod
    def load(
        self, instance: E, only: Iterable[str] | None = None
    ) -> bool | Awaitable[bool]: ...


class MutRepository[E, ID](Repository[E, ID], Protocol):
    @abstractmethod
    def create(
        self, data: Mapping[str, Any], **kwargs
    ) -> Result[E, EntityError] | Awaitable[Result[E, EntityError]]: ...

    @abstractmethod
    def get_or_create(
        self, id_value: ID, data: Mapping[str, Any], **kwargs
    ) -> (
        tuple[Result[E, EntityError], bool] | Awaitable[tuple[Result[E, EntityError], bool]]
    ): ...

    @abstractmethod
    def update_by_id(
        self, id_value: ID, data: Mapping[str, Any], **kwargs
    ) -> Result[None, EntityError] | Awaitable[Result[None, EntityError]]: ...

    @abstractmethod
    def save(
        self, instance: E, **kwargs
    ) -> Result[None, EntityError] | Awaitable[Result[None, EntityError]]: ...

    @abstractmethod
    def save_all(self, objects: Iterable[E], **kwargs) -> int | Awaitable[int]: ...

    @abstractmethod
    def delete(self, instance: E) -> bool | Awaitable[bool]: ...

    @abstractmethod
    def delete_all(self, indices: Iterable[ID] | None = None) -> int | Awaitable[int]: ...

    @abstractmethod
    def delete_by_id(self, id_value: ID) -> bool | Awaitable[bool]: ...


class SyncRepository[E, ID](TypeWrapper[E, ID], ABC):
    @abstractmethod
    def get(self, id_value: ID, **kwargs) -> Maybe[E]: ...

    @abstractmethod
    def get_any(self, indices: Iterable[ID], **kwargs) -> Iterable[Maybe[E]]: ...

    @abstractmethod
    def exists(self, id_value: ID, **kwargs) -> bool: ...

    @abstractmethod
    def load(self, instance: E, only: Iterable[str] | None = None) -> bool: ...


class AsyncRepository[E, ID](TypeWrapper[E, ID], ABC):
    @abstractmethod
    async def get(self, id_value: ID, **kwargs) -> Maybe[E]: ...

    @abstractmethod
    async def get_any(self, indices: Iterable[ID], **kwargs) -> Iterable[Maybe[E]]: ...

    @abstractmethod
    async def exists(self, id_value: ID, **kwargs) -> bool: ...

    @abstractmethod
    async def load(self, instance: E, only: Iterable[str] | None = None) -> bool: ...


class SyncMutRepository[E, ID](SyncRepository[E, ID], ABC):
    @abstractmethod
    def create(self, data: Mapping[str, Any], **kwargs) -> Result[E, EntityError]: ...

    def get_or_create(
        self, id_value: ID, data: Mapping[str, Any], **kwargs
    ) -> tuple[Result[E, EntityError], bool]:
        if instance := self.get(id_value, **kwargs).unwrap_or_none():
            return Ok(instance), False
        return self.create(data, **kwargs), True

    @abstractmethod
    def update_by_id(
        self, id_value: ID, data: Mapping[str, Any], **kwargs
    ) -> Result[None, EntityError]: ...

    @abstractmethod
    def save(self, instance: E, **kwargs) -> Result[None, EntityError]: ...

    @abstractmethod
    def save_all(self, objects: Iterable[E], **kwargs) -> int: ...

    @abstractmethod
    def delete(self, instance: E) -> bool: ...

    @abstractmethod
    def delete_all(self, indices: Iterable[ID] | None = None) -> int: ...

    @abstractmethod
    def delete_by_id(self, id_value: ID) -> bool: ...


class AsyncMutRepository[E, ID](AsyncRepository[E, ID], ABC):
    @abstractmethod
    async def create(self, data: Mapping[str, Any], **kwargs) -> Result[E, EntityError]: ...

    async def get_or_create(
        self, id_value: ID, data: Mapping[str, Any], **kwargs
    ) -> tuple[Result[E, EntityError], bool]:
        if instance := (await self.get(id_value, **kwargs)).unwrap_or_none():
            return Ok(instance), False
        return await self.create(data, **kwargs), True

    @abstractmethod
    async def update_by_id(
        self, id_value: ID, data: Mapping[str, Any], **kwargs
    ) -> Result[None, EntityError]: ...

    @abstractmethod
    async def save(self, instance: E, **kwargs) -> Result[None, EntityError]: ...

    @abstractmethod
    async def save_all(self, objects: Iterable[E], **kwargs) -> int: ...

    @abstractmethod
    async def delete(self, instance: E) -> bool: ...

    @abstractmethod
    async def delete_all(self, indices: Iterable[ID] | None = None) -> int: ...

    @abstractmethod
    async def delete_by_id(self, id_value: ID) -> bool: ...


class SyncFilterable[E, ID](ABC):
    @abstractmethod
    def list(
        self, query: Query | None = None, skip: int = 0, limit: int = 0, **kwargs
    ) -> Iterable[E]: ...

    @abstractmethod
    def sublist(
        self, field: str, query: Query | None = None, skip: int = 0, limit: int = 0, **kwargs
    ) -> Iterable[E]: ...

    @abstractmethod
    def count(self, query: Query | None = None, **kwargs) -> int: ...

    @abstractmethod
    def exists_by(self, query: Query, **kwargs) -> bool: ...


class AsyncFilterable[E, ID](ABC):
    @abstractmethod
    async def list(
        self, query: Query | None = None, skip: int = 0, limit: int = 0, **kwargs
    ) -> Iterable[E]: ...

    @abstractmethod
    async def sublist(
        self, field: str, query: Query | None = None, skip: int = 0, limit: int = 0, **kwargs
    ) -> Iterable[E]: ...

    @abstractmethod
    async def count(self, query: Query | None = None, **kwargs) -> int: ...

    @abstractmethod
    async def exists_by(self, query: Query, **kwargs) -> bool: ...


class SyncMutFilterable[E, ID](SyncFilterable[E, ID], ABC):
    @abstractmethod
    def update_by(self, query: Query, data: Mapping[str, Any], **kwargs) -> int: ...

    @abstractmethod
    def delete_by(self, query: Query, **kwargs) -> int: ...


class AsyncMutFilterable[E, ID](AsyncFilterable[E, ID], ABC):
    @abstractmethod
    async def update_by(self, query: Query, data: Mapping[str, Any], **kwargs) -> int: ...

    @abstractmethod
    async def delete_by(self, query: Query, **kwargs) -> int: ...



from greyhorse.app.abc.providers import SharedProvider
from greyhorse.app.entities.application import Application
from greyhorse.app.schemas.components import ModuleComponentConf, ProvidersConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from .common.functional import FunctionalOperator
from .root import FunctionalOperatorCtrl, FunctionalOperatorService


def __init__():
    return ModuleComponentConf(
        enabled=True,
        path='..module.main',
        provider_imports=[
            ProvidersConf(
                resource=FunctionalOperator, providers=[SharedProvider[FunctionalOperator]]
            )
        ],
        services=[SvcConf(type=FunctionalOperatorService)],
        controllers=[CtrlConf(type=FunctionalOperatorCtrl)],
    )


def test_app() -> None:
    app_conf = __init__()

    app = Application('TestApp')

    res = app.load(app_conf)
    assert res.is_ok()

    res = app.setup()
    assert res.is_ok()

    print('OK')

    prov = app.get_provider(SharedProvider[FunctionalOperator])
    assert prov.is_just()

    prov = prov.unwrap()
    op = prov.borrow().unwrap()

    res = op.add_number(123)
    assert res.is_ok()

    res = op.get_number()
    assert res.is_ok()
    assert res.unwrap() == 123

    res = op.remove_number()
    assert res.is_ok()
    assert res.unwrap() is True

    res = op.get_number()
    assert res.is_err()
    assert res.unwrap_err() == 'Number is not initialized'

    assert app.start()

    # with app.graceful_exit():
    # app.run_sync()
    # await app.run_async()

    assert app.stop()

    res = app.teardown()
    assert res.is_ok()

    res = app.unload()
    assert res.is_ok()



from typing import Any, override

from greyhorse.app.abc.collectors import MutNamedCollector, NamedCollector
from greyhorse.app.abc.controllers import ControllerError
from greyhorse.app.abc.operators import AssignOperator
from greyhorse.app.abc.providers import FactoryError, FactoryProvider
from greyhorse.app.abc.selectors import NamedListSelector, NamedSelector
from greyhorse.app.abc.services import ProvisionError, ServiceError, ServiceState
from greyhorse.app.entities.controllers import SyncController, operator
from greyhorse.app.entities.services import SyncService, provider
from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Err, Ok, Result, do

from ..common.functional import FunctionalOperator, FunctionalOpProvider
from ..common.resources import (
    DictCtxProvider,
    DictMutCtxProvider,
    DictResContext,
    MutDictResContext,
)


class FunctionalOperatorImpl(FunctionalOperator):
    def __init__(self, ctx: DictResContext, mut_ctx: MutDictResContext) -> None:
        self.ctx = ctx
        self.mut_ctx = mut_ctx

    @override
    def add_number(self, value: int) -> Result[None, str]:
        with self.mut_ctx as ctx:
            ctx['number'] = value
            self.mut_ctx.apply()

        return Ok()

    @override
    def get_number(self) -> Result[int, str]:
        with self.ctx as ctx:
            value = Maybe(ctx.get('number'))

        return value.ok_or('Number is not initialized')

    @override
    def remove_number(self) -> Result[bool, str]:
        with self.mut_ctx as ctx:
            value = Maybe(ctx.pop('number', None))
            self.mut_ctx.apply()

        return value.ok_or('Number is not initialized').map(lambda _: True)


class FunctionalOpProviderImpl(FactoryProvider[FunctionalOperator]):
    def __init__(self, ctx_prov: DictCtxProvider, mut_ctx_prov: DictMutCtxProvider) -> None:
        self._ctx_prov = ctx_prov
        self._mut_ctx_prov = mut_ctx_prov

    @override
    def create(self) -> Result[FunctionalOperator, FactoryError]:
        result: Result[tuple[DictResContext, MutDictResContext], str] = do(
            Ok((c1, c2))
            for c1 in self._ctx_prov.borrow()
            for c2 in self._mut_ctx_prov.acquire()
        )

        match result:
            case Ok((ctx, mut_ctx)):
                return Ok(FunctionalOperatorImpl(ctx, mut_ctx))
            case Err(e):
                return FactoryError.Internal(name='FunctionalOperator', details=e).to_result()

        return FactoryError.Internal(
            name='FunctionalOperator', details='Unexpected return'
        ).to_result()

    @override
    def destroy(self, instance: FunctionalOperatorImpl) -> None:
        self._mut_ctx_prov.release(instance.mut_ctx)
        self._ctx_prov.reclaim(instance.ctx)
        del instance


class DictOperatorService(SyncService):
    @override
    def setup(
        self,
        res: Maybe[DictResContext],
        selector: NamedSelector[type, Any],
        list_selector: NamedListSelector[type, Any],
    ) -> Result[ServiceState, ServiceError]:
        if not res:
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        if not selector.has(DictResContext):
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        if not list_selector.has(DictResContext):
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        return super().setup()

    @override
    def teardown(
        self,
        res: Maybe[DictResContext],
        selector: NamedSelector[type, Any],
        list_selector: NamedListSelector[type, Any],
    ) -> Result[ServiceState, ServiceError]:
        if not res:
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        if not selector.has(DictResContext):
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        if not list_selector.has(DictResContext):
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        return super().teardown()

    def start(self) -> None:
        self._switch_to_active(True)

    def stop(self) -> None:
        self._switch_to_active(False)

    @provider(FunctionalOpProvider)
    def create_prov(
        self, ctx_prov: DictCtxProvider, mut_ctx_prov: DictMutCtxProvider
    ) -> Result[FunctionalOpProvider, ProvisionError]:
        return Ok(FunctionalOpProviderImpl(ctx_prov, mut_ctx_prov))


class DictOperatorCtrl(SyncController):
    def __init__(self) -> None:
        super().__init__()
        self._a = Nothing

    def _setter(self, value) -> None:
        self._a = value

    @override
    def setup(self, collector: NamedCollector[type, Any]) -> Result[bool, ControllerError]:
        if not self._a:
            return ControllerError.NoSuchResource(name='DictResContext').to_result()
        collector.add(DictResContext, self._a.unwrap())
        return super().setup(collector)

    @override
    def teardown(
        self, collector: MutNamedCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        if not self._a:
            return ControllerError.NoSuchResource(name='DictResContext').to_result()
        collector.remove(DictResContext, self._a.unwrap())
        return super().teardown(collector)

    @operator(DictResContext)
    def create_op(self):
        return AssignOperator[DictResContext](lambda: self._a, self._setter)



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Mapping
from typing import override

from greyhorse.app.contexts import Context
from greyhorse.app.registries import MutDictRegistry
from greyhorse.maybe import Maybe


class DataStorageEngine(ABC):
    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def active(self) -> bool: ...

    @abstractmethod
    def get_context[T: Context](self, kind: type[T]) -> Maybe[T]: ...

    def start(self) -> Awaitable[None] | None:
        pass

    def stop(self) -> Awaitable[None] | None:
        pass


class DataStorageEngineFactory(ABC):
    @abstractmethod
    def create_engine(self, name: str, *args, **kwargs) -> DataStorageEngine: ...

    @abstractmethod
    def destroy_engine(self, name: str) -> bool: ...

    @abstractmethod
    def get_engine_names(self) -> list[str]: ...

    @abstractmethod
    def get_engine(self, name: str) -> Maybe[DataStorageEngine]: ...

    @abstractmethod
    def get_engines(
        self, names: list[str] | None = None
    ) -> Mapping[str, DataStorageEngine]: ...


class SimpleDataStorageFactory[T: DataStorageEngine](DataStorageEngineFactory, ABC):
    def __init__(self) -> None:
        self._engines = MutDictRegistry[str, DataStorageEngine]()

    @override
    def destroy_engine(self, name: str) -> bool:
        return self._engines.remove(name)

    @override
    def get_engine_names(self) -> list[str]:
        return [k for k, _ in self._engines.items()]

    @override
    def get_engine(self, name: str) -> Maybe[T]:
        return self._engines.get(name)

    @override
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, T]:
        engines = {}

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name).unwrap_or_none():
                engines[name] = engine

        return engines



import asyncio
import signal
import threading
from asyncio import get_running_loop
from collections.abc import Callable, Collection, Generator
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import Any, NoReturn

from greyhorse.app.abc.controllers import Controller
from greyhorse.app.abc.providers import Provider
from greyhorse.app.abc.services import Service, ServiceWaiter
from greyhorse.app.abc.visitor import Visitor
from greyhorse.app.builders.module import load_module, unload_module
from greyhorse.app.entities.components import ModuleComponent
from greyhorse.app.schemas.components import ModuleComponentConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.result import Ok, Result
from greyhorse.utils.invoke import get_asyncio_loop, invoke_sync
from greyhorse.utils.project import get_project_path, get_version


class ApplicationError(Error):
    namespace = 'greyhorse.app.application'

    AlreadyLoaded = ErrorCase(msg='Application already loaded')
    NotLoaded = ErrorCase(msg='Application not loaded')
    Load = ErrorCase(msg='Load error occurred: "{details}"', details=str)
    Unload = ErrorCase(msg='Unload error occurred: "{details}"', details=str)
    Component = ErrorCase(
        msg='Component error in application, details: "{details}"', details=str
    )


class _ElementsVisitor(Visitor):
    def __init__(self, controllers: list[Controller], services: list[Service]) -> None:
        self._controllers: list[Controller] = controllers
        self._services: list[Service] = services

    def visit_controller(self, controller: Controller) -> None:
        self._controllers.append(controller)

    def visit_service(self, service: Service) -> None:
        self._services.append(service)


class Application:
    def __init__(self, name: str, version: str | None = None, debug: bool = False) -> None:
        self._name = name
        self._debug = debug
        self._path = get_project_path().parent
        self._version = version or get_version()
        self._controllers: list[Controller] = []
        self._services: list[Service] = []
        self._root: Maybe[ModuleComponent] = Nothing
        self._conf: Maybe[ModuleComponentConf] = Nothing

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    def get_cwd(self) -> Path:
        return self._path.absolute()

    def add_resource(self, res_type: type, resource: Any, name: str | None = None) -> bool:
        return self._root.map_or(
            False, lambda root: root.add_resource(res_type, resource, name)
        )

    def remove_resource(self, res_type: type, name: str | None = None) -> bool:
        return self._root.map_or(False, lambda root: root.remove_resource(res_type, name=name))

    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]:
        return self._root.and_then(lambda root: root.get_provider(prov_type))

    def load(self, conf: ModuleComponentConf) -> Result[None, ApplicationError]:
        if self._root:
            return ApplicationError.AlreadyLoaded().to_result()

        logger.info('{name}: Application load'.format(name=self.name))

        if not (
            res := load_module(self._name, conf).map_err(
                lambda e: ApplicationError.Load(details=e.message)
            )
        ):
            return res  # type: ignore

        module = res.unwrap()

        try:
            instance = ModuleComponent(
                name=self._name, path=self._name, conf=conf, module=module
            )

        except Exception as e:
            error = ApplicationError.Component(details=str(e))
            logger.error(error.message)
            return error.to_result()

        self._root = Just(instance)
        self._conf = Just(conf)

        logger.info('{name}: Application loaded successfully'.format(name=self.name))
        return Ok()

    def unload(self) -> Result[None, ApplicationError]:
        if not self._root:
            return Ok()

        logger.info('{name}: Application unload'.format(name=self.name))

        conf, self._conf = self._conf.unwrap(), Nothing

        if not (
            res := unload_module(self._name, conf).map_err(
                lambda e: ApplicationError.Unload(details=e.message)
            )
        ):
            return res

        del conf

        logger.info('{name}: Application unloaded successfully'.format(name=self.name))
        return Ok()

    def setup(self) -> Result[None, ApplicationError]:
        return (
            self._root.map(
                lambda root: root.setup()
                .map_err(lambda e: ApplicationError.Component(details=e.message))
                .map(lambda _: root)
            )
            .unwrap_or_else(lambda: ApplicationError.NotLoaded().to_result())
            .map(self._collect_elements)
        )

    def teardown(self) -> Result[None, ApplicationError]:
        self._controllers = []
        self._services = []

        return self._root.map(
            lambda root: root.teardown().map_err(
                lambda e: ApplicationError.Component(details=e.message)
            )
        ).unwrap_or_else(lambda: ApplicationError.NotLoaded().to_result())

    def start(self) -> bool:
        if not self._root:
            return False

        for svc in self._services:
            if hasattr(svc, 'start'):
                invoke_sync(svc.start)

        for ctrl in self._controllers:
            if hasattr(ctrl, 'start'):
                invoke_sync(ctrl.start)

        return True

    def stop(self) -> bool:
        if not self._root:
            return False

        for ctrl in self._controllers:
            if hasattr(ctrl, 'stop'):
                invoke_sync(ctrl.stop)

        for svc in self._services:
            if hasattr(svc, 'stop'):
                invoke_sync(svc.stop)

        return True

    def run_visitor(self, visitor: Visitor) -> bool:
        return self._root.map(lambda root: root.accept_visitor(visitor)).map_or(
            False, lambda _: True
        )

    def run_sync(self, callback: Callable[[], None] | None = None) -> None:
        sync_events: list[threading.Event] = []
        async_events: list[asyncio.Event] = []

        for svc in self._services:
            match svc.waiter:
                case ServiceWaiter.Sync() as event:
                    sync_events.append(event._0)
                case ServiceWaiter.Async() as event:
                    async_events.append(event._0)

        all_events = sync_events + async_events

        async def waiter() -> None:
            async with asyncio.TaskGroup() as tg:
                for e in async_events:
                    tg.create_task(e.wait())

        logger.info('{name}: Application start sync'.format(name=self._name))

        while not all([e.is_set() for e in all_events]):
            if async_events:
                get_running_loop().run_until_complete(waiter())

            sync_events_bools = [e.wait(0.1) for e in sync_events]

            if callback and not all(sync_events_bools):
                callback()

        logger.info('{name}: Application running sync STOPPED'.format(name=self._name))

    async def run_async(self) -> None:
        sync_events: list[threading.Event] = []
        async_events: list[asyncio.Event] = []

        for svc in self._services:
            match svc.waiter:
                case ServiceWaiter.Sync() as event:
                    sync_events.append(event._0)
                case ServiceWaiter.Async() as event:
                    async_events.append(event._0)

        all_events = sync_events + async_events

        logger.info('{name}: Application start async'.format(name=self._name))

        while not all([e.is_set() for e in all_events]):
            async with asyncio.TaskGroup() as tg:
                for se in sync_events:
                    tg.create_task(asyncio.to_thread(se.wait, 0.1))
                for ae in async_events:
                    tg.create_task(ae.wait())

        logger.info('{name}: Application running async STOPPED'.format(name=self._name))

    @contextmanager
    def graceful_exit(
        self, signals: Collection[int] = (signal.SIGINT, signal.SIGTERM)
    ) -> Generator:
        signals = set(signals)
        flag: list[bool] = []

        if loop := get_asyncio_loop():
            for sig_num in signals:
                loop.add_signal_handler(sig_num, self._exit_handler, sig_num, None, flag)
            try:
                yield
            finally:
                for sig_num in signals:
                    loop.remove_signal_handler(sig_num)
        else:
            original_handlers = []

            for sig_num in signals:
                original_handlers.append(signal.getsignal(sig_num))
                signal.signal(sig_num, partial(self._exit_handler, flag=flag))
            try:
                yield
            finally:
                for sig_num, handler in zip(signals, original_handlers, strict=False):
                    signal.signal(sig_num, handler)

    def _collect_elements(self, root: ModuleComponent) -> None:
        visitor = _ElementsVisitor(self._controllers, self._services)
        root.accept_visitor(visitor)

    def _exit_handler(self, sig_num: 'signal.Signals', _, flag: list[bool]) -> None:
        if flag:
            self._second_exit_stage(sig_num)
        else:
            self._first_exit_stage(sig_num)
            flag.append(True)

    def _first_exit_stage(self, sig_num: 'signal.Signals') -> None:
        fail = False

        try:
            self.stop()
        except RuntimeError:
            fail = True

        if fail:
            self._second_exit_stage(sig_num)

    def _second_exit_stage(self, sig_num: 'signal.Signals') -> NoReturn:
        raise SystemExit(128 + sig_num)



import re
from enum import Enum


def capitalize(string: str) -> str:
    return string[0].upper() + string[1:]


def snake2kebab(string: str) -> str:
    return re.sub('_', '-', string)


def kebab2snake(string: str) -> str:
    return re.sub('-', '_', string)


def snake2camel(string: str, upper: bool = False) -> str:
    result = []

    for i, word in enumerate(string.split('_')):
        if upper or i > 0:
            word = word.capitalize()
        result.append(word)

    return ''.join(result)


def prepare_enum_keys(enum_class: type[Enum]):
    return [snake2kebab(str(e.value).lower()) for e in enum_class]


def to_human_size(num: int | float, suffix: str = 'B') -> str:
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return f'{num:3.1f}{unit}{suffix}'
        num /= 1024.0
    return f'{num:.1f}Yi{suffix}'



import asyncio
import inspect
import threading
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any

from greyhorse.enum import Enum, Struct, Tuple, Unit
from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result

from .providers import Provider
from .visitor import Visitor


class ServiceState(Enum):
    Idle = Unit()
    Active = Struct(started=bool)


class ServiceWaiter(Enum):
    Sync = Tuple(threading.Event)
    Async = Tuple(asyncio.Event)


class ServiceError(Error):
    namespace = 'greyhorse.app'

    Factory = ErrorCase(msg='Service factory error: "{details}"', details=str)
    NoSuchResource = ErrorCase(msg='Resource "{name}" is not found', name=str)


type ServiceFactoryFn = Callable[[...], Service | Result[Service, ServiceError]]
type ServiceFactories = dict[type[Service], ServiceFactoryFn]


class ProvisionError(Error):
    namespace = 'greyhorse.app'

    WrongState = ErrorCase(
        msg='Cannot create provider "{name}" because service is in wrong state: "{state}"',
        name=str,
        state=str,
    )

    InsufficientDeps = ErrorCase(
        msg='Cannot create provider "{name}" because dependencies are not enough to satisfy',
        name=str,
    )


@dataclass(slots=True, frozen=True)
class ProviderMember:
    class_name: str
    method_name: str
    resource_type: type
    provider_type: type[Provider]
    method: classmethod
    params: dict[str, type] = field(default_factory=dict)
    ret_type: type | None = None


def _is_provider_member(member: classmethod) -> bool:
    return hasattr(member, '__provider__')


class Service(ABC):
    def __init__(self) -> None:
        self._state = ServiceState.Idle
        self._providers: dict[type, list[type[Provider]]] = defaultdict(list)
        self._provider_members: dict[type[Provider], ProviderMember] = {}
        self._init_provider_members()

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.visit_service(self)

    def _init_provider_members(self) -> None:
        provider_members = inspect.getmembers(self, _is_provider_member)

        for _, member in provider_members:
            member = member.__provider__
            self._providers[member.resource_type].append(member.provider_type)
            self._provider_members[member.provider_type] = member

    def inspect(self, callback: Callable[[ProviderMember], Any]) -> None:
        for member in self._provider_members.values():
            callback(member)

    @property
    def state(self) -> ServiceState:
        return self._state

    @property
    @abstractmethod
    def waiter(self) -> ServiceWaiter: ...

    @abstractmethod
    def setup(
        self, *args, **kwargs
    ) -> Result[ServiceState, ServiceError] | Awaitable[Result[ServiceState, ServiceError]]: ...

    @abstractmethod
    def teardown(
        self, *args, **kwargs
    ) -> Result[ServiceState, ServiceError] | Awaitable[Result[ServiceState, ServiceError]]: ...

    def can_provide(self, resource_type: type) -> bool:
        if resource_type not in self._providers:
            return False
        return len(self._providers[resource_type]) > 0

    def get_resource_types(self) -> list[type]:
        return list(self._providers.keys())

    def get_provider_types(self, resource_type: type) -> list[type[Provider]]:
        if resource_type not in self._providers:
            return []
        return self._providers[resource_type].copy()

    def _switch_to_idle(self) -> None:
        self._state = ServiceState.Idle

    def _switch_to_active(self, started: bool = False) -> None:
        self._state = ServiceState.Active(started=started)



import os
from collections.abc import Mapping
from typing import Any


def default_value(type_: type):
    def getter(value: Any, default: Any):
        return type_(value) if value is not None else default

    return getter


def expandvars_dict(data: Mapping[str, str]) -> Mapping[str, str]:
    """Expands all environment variables in a dictionary."""
    return dict((k, os.path.expandvars(v)) for k, v in data.items())



# mypy: warn_no_return=false,disable_error_code="arg-type,has-type,misc,return-value"
from __future__ import annotations

import inspect
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator, Iterator
from functools import wraps
from typing import TYPE_CHECKING, Any, NoReturn, TypeGuard, TypeVar

from .enum import Enum, Tuple
from .error import Error

if TYPE_CHECKING:
    from .maybe import Maybe


ExcType = TypeVar('ExcType', bound=BaseException)


class Result[T, E](Enum):
    Ok = Tuple(T)
    Err = Tuple(E)

    def __new__(cls, value: T | E | None = None):
        match cls.__name__:
            case 'Result':
                if isinstance(value, Error):
                    return cls.__new_err__(value)
                return cls.__new_ok__(value)
            case 'Ok':
                return cls.__new_ok__(value)
            case 'Err':
                return cls.__new_err__(value)

        raise AssertionError()

    @classmethod
    def __new_ok__(cls, value):
        return super().__new__(Result[type(value), Any].Ok)

    @classmethod
    def __new_err__(cls, value):
        return super().__new__(Result[Any, type(value)].Err)

    def __bool__(self) -> bool:
        return self.is_ok()

    def __hash__(self) -> int:
        match self:
            case Result.Ok(v):
                return hash((True, v))
            case Result.Err(e):
                return hash((False, e))

    def __iter__(self) -> Iterator[T]:
        match self:
            case Result.Ok(v):
                yield v
            case Result.Err(_):

                def _iter() -> Iterator[NoReturn]:
                    # Exception will be raised when the iterator is advanced, not when it's created
                    raise DoException(self)

                return _iter()

    def is_ok(self) -> bool:
        """Returns true if the `Result` is `Ok`."""
        return isinstance(self, Result[T, E].Ok)

    def is_ok_and(self, f: Callable[[T], bool]) -> bool:
        """
        Returns true if the `Result` is `Ok` and the value inside of it matches a predicate.
        """
        match self:
            case Result.Ok(v):
                return f(v)
            case Result.Err(_):
                return False

    def is_err(self) -> bool:
        """Returns true if the `Result` is `Err`."""
        return isinstance(self, Result[T, E].Err)

    def is_err_and(self, f: Callable[[E], bool]) -> bool:
        """
        Returns true if the `Result` is `Err` and the value inside of it matches a predicate.
        """
        match self:
            case Result.Ok(_):
                return False
            case Result.Err(e):
                return f(e)

    def ok(self) -> Maybe[T]:
        """
        Converts the `Result` into a `Maybe[T]` and discarding the error, if any.
        """
        from .maybe import Just, Nothing

        match self:
            case Result.Ok(v):
                return Just(v)
            case Result.Err(_):
                return Nothing

    def err(self) -> Maybe[E]:
        """
        Converts the `Result` into a `Maybe[E]` and discarding the success value, if any.
        """
        from .maybe import Just, Nothing

        match self:
            case Result.Ok(_):
                return Nothing
            case Result.Err(e):
                return Just(e)

    def inspect(self, f: Callable[[T], Any]) -> Result[T, E]:
        """
        Calls a function with the contained value if `Ok`.
        Returns the original `Result`.
        """
        match self:
            case Result.Ok(v):
                f(v)
            case Result.Err(_):
                pass
        return self

    def inspect_err(self, f: Callable[[E], Any]) -> Result[T, E]:
        """
        Calls a function with the contained value if `Err`.
        Returns the original `Result`.
        """
        match self:
            case Result.Ok(_):
                pass
            case Result.Err(e):
                f(e)
        return self

    def expect(self, message: str) -> T:
        """
        Returns the contained `Ok` value if the `Result` is `Ok` or raises a `ResultUnwrapError`.
        """
        match self:
            case Result.Ok(v):
                return v
            case Result.Err(e):
                raise ResultUnwrapError(message, e)

    def expect_err(self, message: str) -> E:
        """
        Returns the contained `Err` value if the `Result` is `Err` or raises a `ResultUnwrapError`.
        """
        match self:
            case Result.Ok(v):
                raise ResultUnwrapError(message, v)
            case Result.Err(e):
                return e

    def unwrap(self) -> T:
        """
        Returns the contained `Ok` value if the `Result` is `Ok` or raises a `ResultUnwrapError`.
        """
        match self:
            case Result.Ok(v):
                return v
            case Result.Err(e):
                raise ResultUnwrapError('Called `Result.unwrap()` on an `Err` value', e)

    def unwrap_err(self) -> E:
        """
        Returns the contained `Err` value if the `Result` is `Err` or raises a `ResultUnwrapError`.
        """
        match self:
            case Result.Ok(v):
                raise ResultUnwrapError('Called `Result.unwrap_err()` on an `Ok` value', v)
            case Result.Err(e):
                return e

    def unwrap_or(self, default: T) -> T:
        """
        Returns the contained `Ok` value or a provided default.
        """
        match self:
            case Result.Ok(v):
                return v
            case Result.Err(_):
                return default

    def unwrap_or_none(self) -> T | None:
        """
        Returns the contained `Ok` value or `None`.
        """
        match self:
            case Result.Ok(v):
                return v
            case Result.Err(_):
                return None

    def unwrap_err_or_none(self) -> T | None:
        """
        Returns the contained `Err` value or `None`.
        """
        match self:
            case Result.Ok(_):
                return None
            case Result.Err(e):
                return e

    def unwrap_or_else(self, f: Callable[[E], T]) -> T:
        """
        Returns the contained `Ok` value or computes it from a closure.
        """
        match self:
            case Result.Ok(v):
                return v
            case Result.Err(e):
                return f(e)

    def unwrap_or_raise(self, exc: Callable[[], ExcType]) -> T:
        """
        Returns the contained `Ok` value or raise the provided exception.
        """
        match self:
            case Result.Ok(v):
                return v
            case Result.Err(e):
                if isinstance(e, BaseException):
                    raise exc() from e
                raise exc()

    def map[U](self, f: Callable[[T], U]) -> Result[U, E]:
        """
        Maps a `Result[T, E]` to `Result[U, E]` by applying a function to a contained `Ok` value,
        leaving an `Err` value untouched.

        This function can be used to compose the results of two functions.
        """
        match self:
            case Result.Ok(v):
                return Result[U, E].Ok(f(v))
            case Result.Err(e):
                return Result[U, E].Err(e)

    async def map_async[U](self, f: Callable[[T], Awaitable[U]]) -> Result[U, E]:
        """
        Maps a `Result[T, E]` to `Result[U, E]` by applying a function to a contained `Ok` value,
        leaving an `Err` value untouched.

        This function can be used to compose the results of two functions.
        """
        match self:
            case Result.Ok(v):
                return Result[U, E].Ok(await f(v))
            case Result.Err(e):
                return Result[U, E].Err(e)

    def map_or[U](self, default: U, f: Callable[[T], U]) -> U:
        """
        Returns the provided default (if `Err`), or applies a function
        to the contained value (if `Ok`).
        """
        match self:
            case Result.Ok(v):
                return f(v)
            case Result.Err(_):
                return default

    def map_or_else[U](self, default_f: Callable[[E], U], f: Callable[[T], U]) -> U:
        """
        Maps a `Result[T, E]` to `U` by applying fallback function `default` to a contained `Err` value,
        or function `f` to a contained `Ok` value.

        This function can be used to unpack a successful result while handling an error.
        """
        match self:
            case Result.Ok(v):
                return f(v)
            case Result.Err(e):
                return default_f(e)

    def map_err[F](self, f: Callable[[E], F]) -> Result[T, F]:
        """
        Maps a `Result[T, E]` to `Result[T, F]` by applying a function to a contained `Err` value,
        leaving an `Ok` value untouched.

        This function can be used to pass through a successful result while handling an error.
        """
        match self:
            case Result.Ok(v):
                return Result[T, F].Ok(v)
            case Result.Err(e):
                return Result[T, F].Err(f(e))

    def and_[U](self, other: Result[U, E]) -> Result[U, E]:
        """
        Returns `other` if the Result is `Ok`, otherwise returns the `Err` value of `self`.
        """
        match self:
            case Result.Ok(_):
                return other
            case Result.Err(_):
                return self

    def and_then[U](self, f: Callable[[T], Result[U, E]]) -> Result[U, E]:
        """
        Calls `f` if the result is `Ok`, otherwise returns the `Err` value of `self`.
        This function can be used for control flow based on `Result` values.
        """
        match self:
            case Result.Ok(v):
                return f(v)
            case Result.Err(_):
                return self

    async def and_then_async[U](
        self, f: Callable[[T], Awaitable[Result[U, E]]]
    ) -> Result[U, E]:
        """
        Calls `f` if the result is `Ok`, otherwise returns the `Err` value of `self`.
        This function can be used for control flow based on `Result` values.
        """
        match self:
            case Result.Ok(v):
                return await f(v)
            case Result.Err(_):
                return self

    def or_[F](self, other: Result[T, F]) -> Result[T, F]:
        """
        Returns `other` if the Result is `Err`, otherwise returns the `Ok` value of `self`.
        """
        match self:
            case Result.Ok(_):
                return self
            case Result.Err(_):
                return other

    def or_else[F](self, f: Callable[[E], Result[T, F]]) -> Result[T, F]:
        """
        Calls `f` if the result is `Err`, otherwise returns the `Ok` value of `self`.
        This function can be used for control flow based on `Result` values.
        """
        match self:
            case Result.Ok(_):
                return self
            case Result.Err(e):
                return f(e)

    def flatten(self) -> Result[T, E]:
        """
        Converts from `Result[Result[T, E], E]` to `Result[T, E]`.
        """
        match self:
            case Result.Ok(v):
                match v:
                    case Result.Ok(_):
                        return v
                    case Result.Err(_):
                        return v
                    case _:
                        return self
            case Result.Err(_):
                return self

    def to_maybe(self) -> Maybe[Result[T, E]]:
        """
        Transposes a `Result` of an `Maybe` into an `Maybe` of a `Result`.

        `Ok(Nothing)` will be mapped to `Nothing`.
        `Ok(Just(_))` and `Err(_)` will be mapped to `Just(Ok(_))` and `Just(Err(_))`.
        """
        from .maybe import Just, Maybe, Nothing

        match self:
            case Result.Ok(v):
                match v:
                    case Maybe.Just(x):
                        return Just(Ok(x))
                    case Maybe.Nothing:
                        return Nothing
                    case _:
                        return Just(self)

            case Result.Err(e):
                return Just(Result[T, E].Err(e))


Ok = Result.Ok
Err = Result.Err


class ResultUnwrapError[V](Exception):
    """
    Exception raised from ``.unwrap_<...>`` and ``.expect_<...>`` calls.

    The contained value can be accessed via the ``.value`` attribute, but
    this is not intended for regular use, as type information is lost:
    ``ResultUnwrapError`` doesn't know about both ``T`` and ``E``, since it's raised
    from ``Ok()`` or ``Err()`` which only knows about either ``T`` or ``E``,
    not both.
    """

    def __init__(self, message: str, value: V) -> None:
        super().__init__(message)
        self._value = value

    @property
    def value(self) -> V:
        """
        Returns the contained value.
        """
        return self._value


class DoException(Exception):
    """
    This is used to signal to `do()` that the result is an `Err`,
    which short-circuits the generator and returns that Err.
    Using this exception for control flow in `do()` allows us
    to simulate `and_then()` in the Err case: namely, we don't call `op`,
    we just return `self` (the Err).
    """

    def __init__[T, E](self, err: Result[T, E].Err) -> None:  # type: ignore
        self.err = err


def is_ok[T, E](result: Result[T, E]) -> TypeGuard[Result[T, E].Ok]:  # type: ignore
    """A typeguard to check if a result is an Ok

    Usage:
    >>> r: Result[int, str] = get_a_result()
    >>> if is_ok(r):
    >>>     r   # r is of type Ok[int]
    >>> elif is_err(r):
    >>>     r   # r is of type Err[str]
    """
    return result.is_ok()


def is_err[T, E](result: Result[T, E]) -> TypeGuard[Result[T, E].Err]:  # type: ignore
    """A typeguard to check if a result is an Err

    Usage:
    >>> r: Result[int, str] = get_a_result()
    >>> if is_ok(r):
    >>>     r   # r is of type Ok[int]
    >>> elif is_err(r):
    >>>     r   # r is of type Err[str]
    """
    return result.is_err()


def as_result_sync[T, E, **P](
    *exceptions: type[E],
) -> Callable[[Callable[P, T]], Callable[P, Result[T, E]]]:
    """
    Make a decorator to turn a function into one that returns a ``Result``.

    Regular return values are turned into ``Ok(return_value)``. Raised
    exceptions of the specified exception type(s) are turned into ``Err(exc)``.
    """
    if not exceptions or not all(
        inspect.isclass(exception) and issubclass(exception, BaseException)
        for exception in exceptions
    ):
        raise TypeError('as_result_sync() requires one or more exception types')

    def decorator(f: Callable[P, T]) -> Callable[P, Result[T, E]]:
        """
        Decorator to turn a function into one that returns a ``Result``.
        """

        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[T, E]:
            try:
                return Ok(f(*args, **kwargs))
            except exceptions as exc:
                return Err(exc)

        return wrapper

    return decorator


def as_result_async[T, E, **P](
    *exceptions: type[E],
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[Result[T, E]]]]:
    """
    Make a decorator to turn an async function into one that returns a ``Result``.
    Regular return values are turned into ``Ok(return_value)``. Raised
    exceptions of the specified exception type(s) are turned into ``Err(exc)``.
    """
    if not exceptions or not all(
        inspect.isclass(exception) and issubclass(exception, BaseException)
        for exception in exceptions
    ):
        raise TypeError('as_result_async() requires one or more exception types')

    def decorator(f: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[Result[T, E]]]:
        """
        Decorator to turn a function into one that returns a ``Result``.
        """

        @wraps(f)
        async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[T, E]:
            try:
                return Ok(await f(*args, **kwargs))
            except exceptions as exc:
                return Err(exc)

        return async_wrapper

    return decorator


def do[T, E](gen: Generator[Result[T, E], None, None]) -> Result[T, E]:
    """Do notation for Result (syntactic sugar for sequence of `and_then()` calls).


    Usage:

    ``` rust
    // This is similar to
    use do_notation::m;
    let final_result = m! {
        x <- Ok("hello");
        y <- Ok(True);
        Ok(len(x) + int(y) + 0.5)
    };
    ```

    ``` rust
    final_result: Result[float, int] = do(
        Ok(len(x) + int(y) + 0.5)
        for x in Ok("hello")
        for y in Ok(True)
    )
    ```

    NOTE: If you exclude the type annotation e.g. `Result[float, int]`
    your type checker might be unable to infer the return type.
    To avoid an error, you might need to help it with the type hint.
    """
    try:
        return next(gen)
    except DoException as e:
        out: Err[E] = e.err  # type: ignore
        return out
    except TypeError as te:
        # Turn this into a more helpful error message.
        # Python has strange rules involving turning generators involving `await`
        # into async generators, so we want to make sure to help the user clearly.
        if "'async_generator' object is not an iterator" in str(te):
            raise TypeError(
                'Got async_generator but expected generator.'
                'See the section on do notation in the README.'
            )
        raise te


async def do_async[T, E](
    gen: Generator[Result[T, E], None, None] | AsyncGenerator[Result[T, E], None],
) -> Result[T, E]:
    """Async version of do. Example:

    ``` python
    final_result: Result[float, int] = await do_async(
        Ok(len(x) + int(y) + z)
        for x in await get_async_result_1()
        for y in await get_async_result_2()
        for z in get_sync_result_3()
    )
    ```

    NOTE: Python makes generators async in a counter-intuitive way.

    ``` python
    # This is a regular generator:
    async def foo(): ...
        do(Ok(1) for x in await foo())
    ```

    ``` python
    # But this is an async generator:
    async def foo(): ...
    async def bar(): ...


    do(Ok(1) for x in await foo() for y in await bar())
    ```

    We let users try to use regular `do()`, which works in some cases
    of awaiting async values. If we hit a case like above, we raise
    an exception telling the user to use `do_async()` instead.
    See `do()`.

    However, for better usability, it's better for `do_async()` to also accept
    regular generators, as you get in the first case:

    ``` python
    async def foo(): ...
        do(Ok(1) for x in await foo())
    ```

    Furthermore, neither mypy nor pyright can infer that the second case is
    actually an async generator, so we cannot annotate `do_async()`
    as accepting only an async generator. This is additional motivation
    to accept either.
    """
    try:
        if isinstance(gen, AsyncGenerator):
            return await gen.__anext__()
        return next(gen)
    except DoException as e:
        out: Err[E] = e.err  # type: ignore
        return out



from typing import Any

from greyhorse.app.abc.providers import MutProvider, SharedProvider
from greyhorse.app.contexts import SyncContext, SyncMutContext

DictResource = dict[str, Any]
DictResContext = SyncContext[DictResource]
MutDictResContext = SyncMutContext[DictResource]

DictCtxProvider = SharedProvider[DictResContext]
DictMutCtxProvider = MutProvider[MutDictResContext]



from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from types import (
    ClassMethodDescriptorType,
    FunctionType,
    GetSetDescriptorType,
    MappingProxyType,
    MemberDescriptorType,
    MethodDescriptorType,
    MethodType,
    MethodWrapperType,
    WrapperDescriptorType,
)
from typing import Any, Protocol


@dataclass
class FieldsCache:
    priv: set[str] = field(default_factory=set)
    pub: set[str] = field(default_factory=set)
    ser: set[str] = field(default_factory=set)


class ModelFieldsMixin(Protocol):
    class Meta:
        private_fields: set[str] = set()
        public_fields: set[str] = set()
        serializable_fields: set[str] = set()
        non_serializable_fields: set[str] = set()
        _fields_cache: dict[str, FieldsCache] = dict()

    @classmethod
    def _fields_cache_key(cls):
        return '.'.join([cls.__module__, cls.__qualname__])

    # noinspection PyProtectedMember
    @classmethod
    def _calculate_fields(cls) -> None:
        private_types = (
            type,
            FunctionType,
            MethodType,
            MappingProxyType,
            WrapperDescriptorType,
            MethodWrapperType,
            MethodDescriptorType,
            ClassMethodDescriptorType,
            GetSetDescriptorType,
            MemberDescriptorType,
        )

        private_fields, public_fields, serializable_fields = set(), set(), set()

        for attr in dir(cls):
            value = getattr(cls, attr)
            if isinstance(value, private_types) or attr[0] == '_':
                private_fields.add(attr)
            elif not isinstance(value, private_types):
                if not isinstance(value, property):
                    serializable_fields.add(attr)
                public_fields.add(attr)

        private_fields.update(cls.Meta.private_fields)
        private_fields.difference_update(cls.Meta.public_fields)
        public_fields.update(cls.Meta.public_fields)
        public_fields.difference_update(cls.Meta.private_fields)
        serializable_fields.update(cls.Meta.serializable_fields)
        serializable_fields.difference_update(cls.Meta.non_serializable_fields)
        cls.Meta._fields_cache[cls._fields_cache_key()] = FieldsCache(
            priv=private_fields, pub=public_fields, ser=serializable_fields
        )

    # noinspection PyProtectedMember
    @classmethod
    def drop_fields_cache(cls) -> None:
        key = cls._fields_cache_key()
        cls.Meta._fields_cache.pop(key, None)

    # noinspection PyProtectedMember
    @classmethod
    def get_private_fields(cls) -> set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].priv

    # noinspection PyProtectedMember
    @classmethod
    def get_public_fields(cls) -> set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].pub

    @classmethod
    def get_fields(cls) -> set[str]:
        return cls.get_public_fields()

    def get_values(
        self,
        only_fields: Sequence[str] | None = None,
        exclude_fields: Sequence[str] | None = None,
    ) -> Mapping[str, Any]:
        public_fields = self.get_public_fields()
        only_fields = set(only_fields) if only_fields else set()
        exclude_fields = set(exclude_fields) if exclude_fields else set()

        return {
            name: getattr(self, name)
            for name in public_fields
            if hasattr(self, name)
            and (not only_fields or name in only_fields)
            and (not exclude_fields or name not in exclude_fields)
        }



import sys

import pydantic

from greyhorse.app.schemas.components import ModuleComponentConf, ModuleConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.result import Ok, Result
from greyhorse.utils.imports import import_path
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync


class ModuleLoadError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    AlreadyLoaded = ErrorCase(msg='Module already loaded: "{path}"', path=str)
    WrongImport = ErrorCase(
        msg='Module import error: "{path}", details: "{details}"', path=str, details=str
    )
    Validation = ErrorCase(
        msg='Module validation error: "{path}", details: "{details}"', path=str, details=str
    )
    Unexpected = ErrorCase(
        msg='Module unexpected error: "{path}", details: "{details}"', path=str, details=str
    )
    InvalidConf = ErrorCase(msg='Module conf is invalid: "{path}"', path=str)


class ModuleUnloadError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    AlreadyUnloaded = ErrorCase(msg='Module already unloaded: "{path}"', path=str)
    Unexpected = ErrorCase(
        msg='Module unexpected error: "{path}", details: "{details}"', path=str, details=str
    )


class ModuleLoader:
    def __init__(self) -> None:
        self._injector = ParamsInjector()

    # noinspection PyProtectedMember
    def load_pass(self, conf: ModuleComponentConf) -> Result[ModuleConf, ModuleLoadError]:
        if conf._conf is not None:
            return ModuleLoadError.AlreadyLoaded(path=conf.path).to_result()

        if not (res := self._load_module(conf)):
            return res

        conf._conf = res.unwrap()

        if not conf._conf.enabled:
            return ModuleLoadError.Disabled(path=conf.path).to_result()

        return Ok(conf._conf)

    # noinspection PyProtectedMember
    def unload_pass(self, conf: ModuleComponentConf) -> Result[None, ModuleUnloadError]:
        if conf._conf is None:
            return ModuleUnloadError.AlreadyUnloaded(path=conf.path).to_result()

        if not conf._conf.enabled:
            return ModuleUnloadError.Disabled(path=conf.path).to_result()

        return self._unload_module(conf)

    # noinspection PyProtectedMember
    @staticmethod
    def _get_module_package(conf: ModuleComponentConf) -> str:
        if not conf.path.startswith('.'):
            return conf.path

        path = conf.path
        dots_count = 0

        for c in path[1:]:
            if c == '.':
                dots_count += 1
            else:
                path = path[dots_count + 1 :]
                break

        init_path = conf._init_path
        dots_count = min(dots_count, len(init_path))
        init_path = init_path[0 : len(init_path) - dots_count]
        return '.'.join([*init_path, path])

    def _load_module(self, conf: ModuleComponentConf) -> Result[ModuleConf, ModuleLoadError]:
        module_path = self._get_module_package(conf)
        logger.info('ModuleLoader: Try to load module "{path}"'.format(path=module_path))

        try:
            func = import_path(f'{module_path}:__init__')

        except (ImportError, AttributeError) as e:
            error = ModuleLoadError.WrongImport(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        injected_args = self._injector(func, values=conf.args)

        try:
            res = invoke_sync(func, *injected_args.args, **injected_args.kwargs)

        except pydantic.ValidationError as e:
            error = ModuleLoadError.Validation(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()
        except Exception as e:
            error = ModuleLoadError.Unexpected(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        if not isinstance(res, ModuleConf):
            error = ModuleLoadError.InvalidConf(path=module_path)
            logger.error(error.message)
            return error.to_result()

        logger.info(
            'ModuleLoader: Module "{path}" loaded successfully'.format(path=module_path)
        )
        return Ok(res)

    # noinspection PyProtectedMember
    def _unload_module(self, conf: ModuleComponentConf) -> Result[None, ModuleUnloadError]:
        module_path = self._get_module_package(conf)
        logger.info('ModuleLoader: Try to unload module "{path}"'.format(path=module_path))

        try:
            func = import_path(f'{module_path}:__fini__')

        except (ImportError, AttributeError):
            pass
        else:
            injected_args = self._injector(func, types={ModuleConf: conf._conf})

            try:
                invoke_sync(func, *injected_args.args, **injected_args.kwargs)

            except Exception as e:
                error = ModuleUnloadError.Unexpected(path=module_path, details=str(e))
                logger.error(error.message)
                return error.to_result()

        conf, conf._conf = conf._conf, None
        del conf
        del sys.modules[module_path]

        logger.info(
            'ModuleLoader: Module "{path}" unloaded successfully'.format(path=module_path)
        )
        return Ok()



from abc import ABC, abstractmethod

from greyhorse.app.abc.providers import FactoryProvider
from greyhorse.result import Result


class FunctionalOperator(ABC):
    @abstractmethod
    def add_number(self, value: int) -> Result[None, str]: ...

    @abstractmethod
    def get_number(self) -> Result[int, str]: ...

    @abstractmethod
    def remove_number(self) -> Result[bool, str]: ...


FunctionalOpProvider = FactoryProvider[FunctionalOperator]
